/**
 * A reproducer for a POST followed by a GET with large headers on the same
 * connection leading to a NULL dereference.
 *
 * @file post-get-crash.c
 */

#include <stdlib.h>
#include <string.h>

#include "curl/curl.h"

static int first_read = 1;

static const char * const DATA = "zzzzzzzzzzz";

/** The POST URL must read the POST body and return a 200 */
static const char * const POST_URL = "https://httpbin.org/post";

/** The GET URL must be hosted by a server that accepts headers 20K in size */
static const char * const GET_URL = "https://httpbin.org/get";

static
size_t curl_request_body_cb(char *dst, size_t size, size_t nmemb, void *userdata)
{
    if (first_read) {
        size_t len = strlen(DATA) - 1;

        first_read = 0;

        memcpy(dst, DATA, len);
        return len;
    }

    return 0;
}

void setup(CURL *handle)
{
    curl_easy_setopt(handle, CURLOPT_VERBOSE, 1);
    curl_easy_setopt(handle, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
    curl_easy_setopt(handle, CURLOPT_POSTFIELDS, NULL);
    curl_easy_setopt(handle, CURLOPT_SSL_VERIFYPEER, 0);
    curl_easy_setopt(handle, CURLOPT_SSL_VERIFYHOST, 0);
}

CURLcode send_post(CURL *handle)
{
    curl_easy_setopt(handle, CURLOPT_URL, POST_URL);
    curl_easy_setopt(handle, CURLOPT_POST, 1);
    curl_easy_setopt(handle, CURLOPT_POSTFIELDSIZE, strlen(DATA)-1);
    curl_easy_setopt(handle, CURLOPT_READFUNCTION, curl_request_body_cb);
    curl_easy_setopt(handle, CURLOPT_READDATA, NULL);
    return curl_easy_perform(handle);
}

CURLcode send_get(CURL *handle)
{
    char large_header_value[3000];
    struct curl_slist *list = NULL;
    CURLcode result;

    curl_easy_setopt(handle, CURLOPT_URL, GET_URL);
    curl_easy_setopt(handle, CURLOPT_POSTFIELDSIZE, 0);
    curl_easy_setopt(handle, CURLOPT_HTTPGET, 1);
    curl_easy_setopt(handle, CURLOPT_READFUNCTION, NULL);
    curl_easy_setopt(handle, CURLOPT_READDATA, NULL);

    large_header_value[0] = 'X';
    large_header_value[1] = ':';
    large_header_value[2] = ' ';
    for (int i = 3; i < 3000-1; i++) {
        large_header_value[i] = 'x';
    }
    large_header_value[3000-1] = '\0';

    for (int i = 0; i < 6; i++) {
        list = curl_slist_append(list, large_header_value);
    }
    curl_easy_setopt(handle, CURLOPT_HTTPHEADER, list);

    result = curl_easy_perform(handle);

    if (list != NULL) {
        curl_slist_free_all(list);
    }
    return result;
}

int main(int argc, char *argv[])
{
    CURLcode result = CURLE_OK;
    CURL *handle = NULL;

    handle = curl_easy_init();
    if (handle == NULL) {
        return EXIT_FAILURE;
    }

    int retcode = EXIT_SUCCESS;
    do {
        setup(handle);

        result = send_post(handle);
        if (result != CURLE_OK) {
            fprintf(stderr, "failed to send POST\n");
            retcode = EXIT_FAILURE;
            break;
        }

        result = send_get(handle);
        if (result != CURLE_OK) {
            fprintf(stderr, "failed to send GET\n");
            retcode = EXIT_FAILURE;
            break;
        }
    }while(0);


    if (handle != NULL) {
        curl_easy_cleanup(handle);
    }

    return retcode;
}
